import inspect
import traceback
from copy import deepcopy
from pprint import pformat
from types import GenericAlias
from typing import get_origin, Annotated

_TOOL_HOOKS = {}
_TOOL_DESCRIPTIONS = {}


def register_tool(func: callable):
    tool_name = func.__name__
    tool_description = inspect.getdoc(func).strip()
    python_params = inspect.signature(func).parameters
    tool_params = []
    for name, param in python_params.items():
        annotation = param.annotation
        if annotation is inspect.Parameter.empty:
            raise TypeError(f"Parameter `{name}` missing type annotation")
        if get_origin(annotation) != Annotated:
            raise TypeError(f"Annotation type for `{name}` must be typing.Annotated")

        typ, (description, required) = annotation.__origin__, annotation.__metadata__
        typ: str = str(typ) if isinstance(typ, GenericAlias) else typ.__name__
        if not isinstance(description, str):
            raise TypeError(f"Description for `{name}` must be a string")
        if not isinstance(required, bool):
            raise TypeError(f"Required for `{name}` must be a bool")

        tool_params.append({
            "name": name,
            "description": description,
            "type": typ,
            "required": required
        })
    tool_def = {
        "name": tool_name,
        "description": tool_description,
        "params": tool_params
    }

    # print("[registered tool] " + pformat(tool_def))
    _TOOL_HOOKS[tool_name] = func
    _TOOL_DESCRIPTIONS[tool_name] = tool_def

    return func


def dispatch_tool(tool_name: str, tool_params: dict) -> str:
    if tool_name not in _TOOL_HOOKS:
        return f"Tool `{tool_name}` not found. Please use a provided tool."
    tool_call = _TOOL_HOOKS[tool_name]
    try:
        ret = tool_call(**tool_params)
    except:
        ret = traceback.format_exc()
    return str(ret)


def get_tools() -> dict:
    return deepcopy(_TOOL_DESCRIPTIONS)

# Tool Definitions

import re
@register_tool
def robot_orders_generator(
        description: Annotated[str, '描述机器人需要执行的动作，如 "向前直行15米然后左转90度"。', True],
) -> str:
    """
    机器人指令生成器：这个工具根据描述中的字符逐个生成机器人执行的控制指令，并区分距离和角度。
    """
    code = ""
    i = 0
    while i < len(description):
        if description[i] in ["前", "直"]:
            i += 1
            distance_match = re.search(r'\d+', description[i:])
            if distance_match:
                distance = distance_match.group(0)
                code += f"engi_mov_({distance})\n"
                i += distance_match.end()
        elif description[i] in ["左", "右"]:
            direction = "left" if description[i] == "左" else "right"
            i += 1
            angle_match = re.search(r'\d+', description[i:])
            if angle_match:
                angle = angle_match.group(0)
                i += angle_match.end()
                if i < len(description) and description[i] == "度":
                    i += 1
                    code += f"engi_turn_{direction}({angle})\n"
                else:
                    code += f"engi_turn_{direction}(90)\nmove_forward({angle})\n"
        else:
            i += 1

    return code

if __name__ == "__main__":
    # print(dispatch_tool("get_weather", {"city_name": "beijing"}))
    print(get_tools())
